/*
	ARTK_MMD_TWINS
	Copyright (C) 2014  I.Kato @ BS21 Lab

	original:
		ARTK_MMD
		Copyright (C) 2009  PY

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifdef	_WIN32
#include	<windows.h>
#include	<locale.h>
#endif

#include	<gl/glut.h>
#include	<stdio.h>

#include	"ARTK/ARTK.h"
#include	"BulletPhysics/BulletPhysics.h"

#include	"MMD/PMDModel.h"
#include	"MMD/VMDMotion.h"

#include	"SaveImage.h"


extern cBulletPhysics	g_clBulletPhysics;

static cPMDModel	g_clPMDModel;
static cPMDModel	g_clPMDModel2;  // ̖
static cVMDMotion	g_clVMDMotion;
static cVMDMotion	g_clVMDMotion2;  // ̖
static bool			g_bMotionPause;   // [V/~
//static bool		g_bBulltPause;    // Z̓/~
static bool			g_bDispFPS;       // t[[g\
static bool			g_bDispAxis;      // XYZ\
static bool			g_bDispShadow;    // e\
static bool			g_bDispMatrix;    // s\
static bool			g_bDispWireframe; // Ct[\
static bool			g_bDispModel;  // ̖ڂ\邩ǂ
static bool			g_bDispTwin;   // ̖ڂ\邩ǂ
static Matrix		g_matPlanarProjection; // ep̕ʓes
static float		g_fShiftX;  // ̖ڂ̃VtgX
static float		g_fShiftY;  // ̖ڂ̃VtgY
static float		g_fShiftZ;  // ̖ڂ̃VtgZ
static double		g_dMatProjection[16]; // ˉes
static double		g_dMatModelView [16]; // fr[s

char				g_szBaseDir[260]; // fBNg

static void init( void );
static void cleanup( void );
static void display( void );
static void resize( int w, int h );
static void visibility( int visible );
static void idle( void );
static float getElapsedFrame( void );
static void keyboard( unsigned char key, int x, int y );
static void keyboard_sp( int key, int x, int y ); // ㉺EL[̏
static void menu( int value );
static void openNewModel1( void );
static void openNewModel2( void );
static void openNewMotion( void );
static void restartMotion( void ); // ŏ烂[V蒼
static void calcFps( void );
static void dispMatModelView( void );
static void drawString( const char *szStr, int iPosX, int iPosY );


//==========
// main֐
//==========
int main( int argc, char *argv[] )
{
#if 0
	for (int i = 0; i < argc; i++) {
		MessageBoxA( NULL, argv[i], "argv", 0 );
	}
#else
	if (argc >= 1) {
		strcpy (g_szBaseDir, argv[0]); // ApppX
		char *ptr = strrchr (g_szBaseDir, '\\');
		if (ptr == NULL) {
			ptr = strrchr (g_szBaseDir, '/');
			if (ptr == NULL) {
				MessageBoxA( NULL, "path error", "main", 0 );
				exit( -1 );
			}
		}
		*ptr = '\0';
//		MessageBoxA( NULL, g_szBaseDir, "main", 0 );
	}
#endif

	// GLUT̏
	glutInit( &argc, argv );

	// ARToolKit̏
	if( !g_clARTK.initialize() )
	{
		MessageBoxA( NULL, "g_clARTK.initialize() error", "main", 0 );
		exit( -2 );
	}

	// `EBhE̍쐬
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL );
//	glutInitWindowSize( 640, 480 );
	glutInitWindowSize( g_clARTK.getCamImgSizeX(), g_clARTK.getCamImgSizeY() ); // Jɍ킹
	glutCreateWindow( "ARTK MMD TWINS" );

	// argl̏
	if( !g_clARTK.initArgl() )
	{
		MessageBoxA( NULL, "g_clARTK.initArgl() error", "main", 0 );
		exit( -3 );
	}

	// OpenGL ext@\̏: glutCreateWindow()̌ɂKv
	Init_OpenGL_ExtFunc();

	// GLUT Ăяo郆[U[֐̐ݒ
	glutDisplayFunc   ( display );
	glutReshapeFunc   ( resize );
	glutVisibilityFunc( visibility );
	glutKeyboardFunc  ( keyboard );
	glutSpecialFunc   ( keyboard_sp ); // ㉺EL[̏
	atexit( cleanup );

    // Popup Menu ̏
	glutCreateMenu( menu );
		glutAddMenuEntry( "Open Model 1 (PMD/PMX)  [O]",  1 );
		glutAddMenuEntry( "Open Model 2 (PMD/PMX)  [P]",  2 );
		glutAddMenuEntry( "Open Motion (VMD)       [M]",  3 );
		glutAddMenuEntry( "Restart Motion          [R]",  4 );
		glutAddMenuEntry( "Look at me!    (On/Off) [L]",  5 );
		glutAddMenuEntry( "Display Twin   (On/Off) [T]",  6 );
		glutAddMenuEntry( "Display FPS    (On/Off) [F]",  7 );
		glutAddMenuEntry( "Display Axis   (On/Off) [X]",  8 );
		glutAddMenuEntry( "Display Bin    (On/Off) [B]",  9 );
		glutAddMenuEntry( "Display Shadow (On/Off) [H]", 10 );
		glutAddMenuEntry( "Display Matrix (On/Off) [A]", 11 );
		glutAddMenuEntry( "Wire frame     (On/Off) [W]", 12 );
		glutAddMenuEntry( "Save to PNG file        [S]", 13 );
	glutAttachMenu( GLUT_RIGHT_BUTTON );

	init();

	glutMainLoop();

	return 0;
}

//--------
// 
//--------
void init( void )
{
	glClearColor( 0.0f, 0.0f, 1.0f, 1.0f ); // 
//	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f ); // 
	glClearStencil( 0 );
	glEnable( GL_DEPTH_TEST );

	glEnable( GL_NORMALIZE );

	glEnable ( GL_TEXTURE_2D );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

	glEnable  ( GL_CULL_FACE );
	glCullFace( GL_FRONT );

	glEnable   ( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ); // At@Euh

	glEnable   ( GL_ALPHA_TEST );
	glAlphaFunc( GL_GEQUAL, 0.05f ); // 0.05 ȏ̃At@ŕ`

	// 0̐ݒ
	const float	fLightPos[] = { 0.45f, 0.55f, 1.0f, 0.0f }; // WnFw=0.0Ȃ疳
	const float	fLightDif[] = { 0.9f,  0.9f,  0.9f, 1.0f };
	const float	fLightAmb[] = { 0.9f,  0.9f,  0.9f, 1.0f };

	glLightfv( GL_LIGHT0, GL_POSITION, fLightPos ); // ̈ʒu
	glLightfv( GL_LIGHT0, GL_DIFFUSE,  fLightDif ); // gU
	glLightfv( GL_LIGHT0, GL_AMBIENT,  fLightAmb ); // 
	glLightfv( GL_LIGHT0, GL_SPECULAR, fLightAmb ); // ʌ
	glEnable ( GL_LIGHT0 );

	glEnable( GL_LIGHTING );

	// ep̕ʓes̍쐬
	Vector4		vec4Plane = { 0.0f, 1.0f, 0.0f, 0.0f };		// { a, b, c, d }  ax + by + cz + d = 0 (eʂ̕)
	Vector3		vec3LightPos = { 10.0f, 70.0f, -20.0f };	// Cg̈ʒu

	MatrixPlanarProjection( g_matPlanarProjection, &vec4Plane, &vec3LightPos );

	// ZBullet ̏
	g_clBulletPhysics.initialize();

	g_bMotionPause   = false;  // [V/~
//	g_bBulltPause    = false;  // Z̓/~
	g_bDispFPS       = false;
	g_bDispAxis      = false;  // XYZ\
	g_bDispShadow    = true;   // e\
	g_bDispMatrix    = false;
	g_bDispWireframe = false;  // Ct[\
	g_bDispModel     = false;
	g_bDispTwin      = false;
	g_fShiftX = 15.0;  // ̖ڂ̃VtgX̏l
	g_fShiftY =  0.0;  // ̖ڂ̃VtgY̏l
	g_fShiftZ =  0.0;  // ̖ڂ̃VtgZ̏l

	// ̖ڂ̏
	g_clVMDMotion.loadDefFace();

	// ̖ڂ̏
	g_clVMDMotion2.loadDefFace();
}

//------------------
// vOI
//------------------
void cleanup( void )
{
	g_clPMDModel.release();
	g_clPMDModel2.release();
	g_clBulletPhysics.release();
}

//----------
// \֐
//----------
void display( void )
{
	glClear( GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glDisable( GL_CULL_FACE );
	glDisable( GL_ALPHA_TEST );
	glDisable( GL_BLEND );

	g_clARTK.display();

	glEnable( GL_CULL_FACE );
	glEnable( GL_ALPHA_TEST );
	glEnable( GL_BLEND );

	if( g_clARTK.isMarkerFound() )
	{
		glMatrixMode( GL_PROJECTION );
		g_clARTK.getProjectionMat( g_dMatProjection );
		glLoadMatrixd( g_dMatProjection ); // ˉes(En)

		glMatrixMode( GL_MODELVIEW );
		glPushMatrix();

//			glLoadIdentity();
			g_clARTK.getModelViewMat( g_dMatModelView ); // ARToolkit ̃fr[s(En)̎擾
			glLoadMatrixd( g_dMatModelView );

			if (g_bDispAxis) {  // XYZ\
				glScalef( 1.0f, 1.0f, 1.0f );
				glEnable ( GL_DEPTH_TEST ); // ɂƁAႦf̎Oɂ낤ƁAXYZ̐BĂ܂
				glDisable( GL_LIGHTING );   // LɂƁAXYZƂDFŕ`悳Ă܂
				glDisable( GL_TEXTURE_2D ); // LɂƁAXYZƂDFŕ`悳Ă܂
				glBegin( GL_LINES );
					glColor4f(  1.0f, 0.0f, 0.0f, 1.0f ); // 
					glVertex3f(  0.0f,  0.0f,  0.0f );
					glVertex3f( 10.0f,  0.0f,  0.0f ); // X=10.0

					glColor4f(  0.0f, 1.0f, 0.0f, 1.0f ); // 
					glVertex3f(  0.0f,  0.0f,  0.0f );
					glVertex3f(  0.0f, 10.0f,  0.0f ); // Y=10.0

					glColor4f(  0.0f, 0.0f, 1.0f, 1.0f ); // 
					glVertex3f(  0.0f, 0.0f,  0.0f );
					glVertex3f(  0.0f, 0.0f, 10.0f ); // Z=10.0
				glEnd();
				glEnable( GL_LIGHTING );
			}

			glScalef( 1.0f, 1.0f, -1.0f );	// n  En  sĂȂpmd/pmxf(n)wɂȂ
			glCullFace(GL_FRONT); // \ʂ`悷lɂ

			glPolygonMode( GL_FRONT_AND_BACK, g_bDispWireframe ? GL_LINE : GL_FILL ); // C[t[ON/OFF

			// ̖ڂ̏
			if (g_bDispModel) {
				g_clPMDModel.render();
			}

			// ̖ڂ̏
			if (g_bDispTwin) {
				glTranslatef ( -g_fShiftX, -g_fShiftY, -g_fShiftZ );  // X/Y/Z ɂ炷
				g_clPMDModel2.render();
				glTranslatef (  g_fShiftX,  g_fShiftY,  g_fShiftZ );  // ɖ߂
			}

			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL ); // C[t[OFF

			if (g_bDispShadow) {  // e\
				// XeVobt@ɉě``
				glDisable( GL_CULL_FACE );
				glDisable( GL_TEXTURE_2D );
				glDisable( GL_LIGHTING );

				glEnable( GL_STENCIL_TEST );
				glStencilFunc( GL_ALWAYS, 1, ~0 );
				glStencilOp( GL_REPLACE, GL_KEEP, GL_REPLACE );

				glColorMask( 0, 0, 0, 0 );
				glDepthMask( 0 );

					glMultMatrixf( (const float *)g_matPlanarProjection ); // ep̕ʓes

					// ̖ڂ̏
					if (g_bDispModel) {
						g_clPMDModel.renderForShadow();		// ep̕`
					}

					// ̖ڂ̏
					if (g_bDispTwin) {
						glTranslatef ( -g_fShiftX, -g_fShiftY, -g_fShiftZ );  // X/Y/Z ɂ炷
						g_clPMDModel2.renderForShadow();		// ep̕`
						glTranslatef (  g_fShiftX,  g_fShiftY,  g_fShiftZ );  // ɖ߂
					}

				glColorMask( 1, 1, 1, 1 );

				// XeVobt@̉ě`hԂ
				float	fWndW = glutGet( GLUT_WINDOW_WIDTH ),
						fWndH = glutGet( GLUT_WINDOW_HEIGHT );

				glStencilFunc( GL_EQUAL, 1, ~0);
				glStencilOp( GL_KEEP, GL_KEEP ,GL_KEEP );

				glDisable( GL_DEPTH_TEST );

				glMatrixMode( GL_PROJECTION );
				glPushMatrix();
					glLoadIdentity();
					gluOrtho2D( 0.0f, fWndW, 0.0f, fWndH );

					glMatrixMode( GL_MODELVIEW );
					glPushMatrix();
						glLoadIdentity();

							glColor4f( 0.2f, 0.2f, 0.2f, 0.5f );

							glBegin( GL_TRIANGLE_FAN );
								glVertex2f(  0.0f, fWndH );
								glVertex2f( fWndW, fWndH );
								glVertex2f( fWndW,  0.0f );
								glVertex2f(  0.0f,  0.0f );
							glEnd();

						glMatrixMode( GL_PROJECTION );
					glPopMatrix();
					glMatrixMode( GL_MODELVIEW );
				glPopMatrix();

				glDepthMask( 1 );
				glEnable ( GL_LIGHTING );
				glEnable ( GL_DEPTH_TEST );
				glDisable( GL_STENCIL_TEST );

			} else { // g_bDispShadow e\

				glScalef( 1.0f, -1.0f, 1.0f );	// pmd/pmxf(n)  Enɕϊ { XZʂɋ
				glCullFace(GL_BACK); // ʂ`悷lɂBȂƃf`悪ُɂȂ

				glPolygonMode( GL_FRONT_AND_BACK, g_bDispWireframe ? GL_LINE : GL_FILL ); // C[t[ON/OFF

				// ̖ڂ̏
				if (g_bDispModel) {
					g_clPMDModel.render();
				}

				// ̖ڂ̏
				if (g_bDispTwin) {
					glTranslatef ( -g_fShiftX, -g_fShiftY, -g_fShiftZ );  // X/Y/Z ɂ炷
					g_clPMDModel2.render();
					glTranslatef (  g_fShiftX,  g_fShiftY,  g_fShiftZ );  // ɖ߂
				}

				glPolygonMode( GL_FRONT_AND_BACK, GL_FILL ); // C[t[OFF

				glCullFace(GL_FRONT); // \ʂ`悷lɂ
			} // g_bDispShadow e\

		glPopMatrix();
	}

	calcFps();

	dispMatModelView();

	// _uEobt@ swap
	glutSwapBuffers();
}

//--------------------------------
// EChETCYꂽƂ
//--------------------------------
void resize( int w, int h )
{
#if 0
	// AXyNg䂪4F3ŌŒ
	int		iResizeW = (float)h * 4.0f / 3.0f;
	int		iResizeH = (float)w * 3.0f / 4.0f;
#else
	// AXyNgJɍ킹
	int		iResizeW = (float)h * g_clARTK.getCamImgSizeX() / g_clARTK.getCamImgSizeY();
	int		iResizeH = (float)w * g_clARTK.getCamImgSizeY() / g_clARTK.getCamImgSizeX();
#endif

	if( iResizeW < w )
		iResizeH = h;
	else
		iResizeW = w;

	glutReshapeWindow( iResizeW, iResizeH );

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );
	glViewport( 0, 0, iResizeW, iResizeH );
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

//----------------------------------------
// EChẼrWreBωƂ
//----------------------------------------
void visibility( int visible )
{
	if( visible == GLUT_VISIBLE )	glutIdleFunc( idle );
	else							glutIdleFunc( NULL );
}

//------------
// ACh
//------------
void idle( void )
{
	g_clARTK.update();

	Vector3	vecCamPos;
	g_clARTK.getCameraPos( &vecCamPos );

	float		fElapsedFrame = getElapsedFrame();

	if( fElapsedFrame > 10.0f )	fElapsedFrame = 10.0f;

	if (g_bMotionPause) { // [V~
		fElapsedFrame = 0.0f;
	}

	if (g_bDispModel) {
		g_clPMDModel.updateMotion( fElapsedFrame ); // ZÕ{[ό`
		g_clPMDModel.updateNeckBone( &vecCamPos );
	}

	if (g_bDispTwin) { // ̖ڂ̏
		g_clPMDModel2.updateMotion( fElapsedFrame ); // ZÕ{[ό`
		g_clPMDModel2.updateNeckBone( &vecCamPos );
	}

//	if (g_bBulltPause == false) {   // Z̓/~  ~邾ł͑ʖڂȖ͗lBGW̕𒼐ڑ삷KvH
		g_clBulletPhysics.update( fElapsedFrame );
//	}

	if (g_bDispModel) {
		g_clPMDModel.updateMotionAfterBullet( fElapsedFrame ); // Z̃{[ό`
	}

	if (g_bDispTwin) { // ̖ڂ̏
		g_clPMDModel2.updateMotionAfterBullet( fElapsedFrame ); // Z̃{[ό`
	}

	if (g_bDispModel) {
		g_clPMDModel.updateSkinning();
	}
	if (g_bDispTwin) { // ̖ڂ̏
		g_clPMDModel2.updateSkinning();
	}

	glutPostRedisplay();
}

//-----------------------------------------------
// o߃t[Ԃ(1.0 = 1/30s = 33.333ms)
//-----------------------------------------------
float getElapsedFrame( void )
{
	static int	s_iPrevTime = 0;
	int			iTime = glutGet( GLUT_ELAPSED_TIME );
	float		fDiffTime;

	if( s_iPrevTime == 0 )	s_iPrevTime = iTime;
	fDiffTime = (float)(iTime - s_iPrevTime) * (30.0f / 1000.0f);

	s_iPrevTime = iTime;

	return fDiffTime;
}

//----------------------------
// L[{[h͂Ƃ
//----------------------------
void keyboard( unsigned char key, int x, int y )
{
	switch( key )
	{
		case 'o':	// f1Ǎ
		case 'O':
			openNewModel1();
			break;

		case 'p':	// f2Ǎ
		case 'P':
			openNewModel2();
			break;

		case 'm':	// [VǍ
		case 'M':
			openNewMotion();
			break;

		case ' ':	// [V/~
			g_bMotionPause = !g_bMotionPause;
			break;

//		case 'b':  // Z̓/~
//		case 'B':
//			g_bBulltPause = !g_bBulltPause ;
//			break;

		case 'r': // ŏ烂[V蒼
		case 'R':
			restartMotion();
			break;

		case 'l':	// Jڐ[hOn/Off
		case 'L':
			g_clPMDModel .toggleLookAtFlag();
			g_clPMDModel2.toggleLookAtFlag(); // ̖ڂ̏
			break;

		case 't':	// ̖ڂ̕\On/Off
		case 'T':
			g_bDispTwin = !g_bDispTwin;
			break;

		case 'f':	// t[[g\On/Off
		case 'F':
			g_bDispFPS = !g_bDispFPS;
			break;

		case 'x':	// XYZ\On/Off
		case 'X':
			g_bDispAxis = !g_bDispAxis;
			break;

		case 'b':	// l摜\On/Off
		case 'B':
			arDebug = !arDebug; // ARToolkit̃fobOtOgO
			break;

		case 'h':	// e\On/Off
		case 'H':
			g_bDispShadow = !g_bDispShadow;
			break;

		case 'a':	// s\On/Off
		case 'A':
			g_bDispMatrix = !g_bDispMatrix;
			break;

		case 'w':	// Ct[\On/Off
		case 'W':
			g_bDispWireframe = !g_bDispWireframe;
			break;

		case '+':	// fg
			g_clARTK.addViewScaleFactor( -0.01 );
			break;

		case '-':	// fk
			g_clARTK.addViewScaleFactor(  0.01 );
			break;

		case 's':	// 摜ۑ
		case 'S':
			SaveImageToPng();
			break;

	}
}

//----------------------------
// ȃL[{[h͂Ƃ
//----------------------------
void keyboard_sp( int key, int x, int y )
{
	switch( key )
	{

		case GLUT_KEY_RIGHT: // lڂ̃VtgX
			g_fShiftX += 1.0;
			break;

		case GLUT_KEY_LEFT: // lڂ̃VtgX
			g_fShiftX -= 1.0;
			break;

		case GLUT_KEY_PAGE_DOWN: // lڂ̃VtgY
			g_fShiftY += 1.0;
			break;

		case GLUT_KEY_PAGE_UP: // lڂ̃VtgY
			g_fShiftY -= 1.0;
			break;

		case GLUT_KEY_UP: // lڂ̃VtgZ
			g_fShiftZ += 1.0;
			break;

		case GLUT_KEY_DOWN: // lڂ̃VtgZ
			g_fShiftZ -= 1.0;
			break;

		case GLUT_KEY_HOME: // lڂ̃VtgX/Z lɖ߂
			g_fShiftX = 15.0;  // ̖ڂ̃VtgX ̏l
			g_fShiftY =  0.0;  // ̖ڂ̃VtgY ̏l
			g_fShiftZ =  0.0;  // ̖ڂ̃VtgZ ̏l
			break;

	}
}

//--------------------------------
// j[̍ڂIꂽƂ
//--------------------------------
void menu( int value )
{
	switch( value )
	{
		case 1 :	// f1Ǎ
			openNewModel1();
			break;

		case 2 :	// f2Ǎ
			openNewModel2();
			break;

		case 3 :	// [VǍ
			openNewMotion();
			break;

		case 4: // ŏ烂[V蒼
			restartMotion();
			break;

		case 5 :	// Jڐ[hOn/Off
			g_clPMDModel .toggleLookAtFlag();
			g_clPMDModel2.toggleLookAtFlag(); // ̖ڂ̏
			break;

		case 6:	// ̖ڂ̕\On/Off
			g_bDispTwin = !g_bDispTwin;
			break;

		case 7:	// t[[g\On/Off
			g_bDispFPS = !g_bDispFPS;
			break;

		case 8:	// XYZ\On/Off
			g_bDispAxis = !g_bDispAxis;
			break;

		case 9:	// l摜\On/Off
			arDebug = !arDebug; // ARToolkit̃fobOtOgO
			break;

		case 10:	// e\On/Off
			g_bDispShadow = !g_bDispShadow;
			break;

		case 11:	// s\On/Off
			g_bDispMatrix = !g_bDispMatrix;
			break;

		case 12:	// Ct[\On/Off
			g_bDispWireframe = !g_bDispWireframe;
			break;

		case 13:	// 摜ۑ
			SaveImageToPng();
			break;

//		case 14:  // Z̓/~
//			g_bBulltPause = !g_bBulltPause ;
//			break;

	}
}

//----------------------
// f1t@CJ
//----------------------
void openNewModel1( void )
{
#ifdef	_WIN32
	OPENFILENAME	ofn;
	wchar_t			wszFileName[256],
					wszFile[64];
	char			szFileName[256];

	ZeroMemory( &wszFileName, 256 );
	ZeroMemory( &wszFile,      64 );
	ZeroMemory( &ofn,    sizeof(OPENFILENAME) );
	ofn.lStructSize    = sizeof(OPENFILENAME);
	ofn.hwndOwner      = NULL;
//	ofn.lpstrFilter    = L"PMD File(*.pmd)\0*.pmd\0\0";
	ofn.lpstrFilter    = L"PMD File(*.pmd)|PMX File(*.pmx)\0*.pmd;*.pmx\0\0";
	ofn.lpstrFile      = wszFileName;
	ofn.lpstrFileTitle = wszFile;
	ofn.nMaxFile       = MAX_PATH;
	ofn.nMaxFileTitle  = sizeof(wszFile);
	ofn.Flags          = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
	ofn.lpstrDefExt    = L"pmd";
	ofn.lpstrTitle     = L"Open";

	if( GetOpenFileName( &ofn ) )
	{
		SetCurrentDirectory( wszFileName );

		setlocale( LC_ALL, "Japanese_Japan.932" );
	    wcstombs( szFileName, wszFileName, 255 );

		// ̖ڂ̏
		g_bDispModel = g_clPMDModel.load( szFileName );

		if (g_bDispModel) {
			g_clPMDModel.setMotion( &g_clVMDMotion, false );
//			g_clPMDModel.updateMotion( 0.0f );
			g_clPMDModel.resetRigidBodyPos();
		}
	}
#endif
}

//----------------------
// f2t@CJ
//----------------------
void openNewModel2( void )
{
#ifdef	_WIN32
	OPENFILENAME	ofn;
	wchar_t			wszFileName[256],
					wszFile[64];
	char			szFileName[256];

	ZeroMemory( &wszFileName, 256 );
	ZeroMemory( &wszFile,      64 );
	ZeroMemory( &ofn,    sizeof(OPENFILENAME) );
	ofn.lStructSize    = sizeof(OPENFILENAME);
	ofn.hwndOwner      = NULL;
//	ofn.lpstrFilter    = L"PMD File(*.pmd)\0*.pmd\0\0";
	ofn.lpstrFilter    = L"PMD File(*.pmd)|PMX File(*.pmx)\0*.pmd;*.pmx\0\0";
	ofn.lpstrFile      = wszFileName;
	ofn.lpstrFileTitle = wszFile;
	ofn.nMaxFile       = MAX_PATH;
	ofn.nMaxFileTitle  = sizeof(wszFile);
	ofn.Flags          = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
	ofn.lpstrDefExt    = L"pmd";
	ofn.lpstrTitle     = L"Open";

	if( GetOpenFileName( &ofn ) )
	{
		SetCurrentDirectory( wszFileName );

		setlocale( LC_ALL, "Japanese_Japan.932" );
	    wcstombs( szFileName, wszFileName, 255 );

		// ̖ڂ̏
		g_bDispTwin = g_clPMDModel2.load( szFileName );
		g_fShiftX   = 15.0;  // ̖ڂ̃VtgX ̏l
		g_fShiftY   =  0.0;  // ̖ڂ̃VtgY ̏l
		g_fShiftZ   =  0.0;  // ̖ڂ̃VtgZ ̏l

		if (g_bDispTwin) {
			g_clPMDModel2.setMotion( &g_clVMDMotion2, false );
//			g_clPMDModel2.updateMotion( 0.0f );
			g_clPMDModel2.resetRigidBodyPos();
		}
	}
#endif
}

//--------------------------
// [Vt@CJ
//--------------------------
void openNewMotion( void )
{
#ifdef	_WIN32
	OPENFILENAME	ofn;
	wchar_t			wszFileName[256],
					wszFile[64];
	char			szFileName[256];

	ZeroMemory( &wszFileName, 256 );
	ZeroMemory( &wszFile,      64 );
	ZeroMemory( &ofn,    sizeof(OPENFILENAME) );
	ofn.lStructSize    = sizeof(OPENFILENAME);
	ofn.hwndOwner      = NULL;
	ofn.lpstrFilter    = L"VMD File(*.vmd)\0*.vmd\0\0";
	ofn.lpstrFile      = wszFileName;
	ofn.lpstrFileTitle = wszFile;
	ofn.nMaxFile       = MAX_PATH;
	ofn.nMaxFileTitle  = sizeof(wszFile);
	ofn.Flags          = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
	ofn.lpstrDefExt    = L"vmd";
	ofn.lpstrTitle     = L"Open";

	if( GetOpenFileName( &ofn ) )
	{
		SetCurrentDirectory( wszFileName );

		setlocale( LC_ALL, "Japanese_Japan.932" );
	    wcstombs( szFileName, wszFileName, 255 );

		// ̖ڂ̏
		g_clVMDMotion.load( szFileName );

		if (g_bDispModel) {
			g_clPMDModel.setMotion( &g_clVMDMotion, false );
			g_clPMDModel.updateMotion( 0.0f );
			g_clPMDModel.resetRigidBodyPos();
		}

		// ̖ڂ̏
		g_clVMDMotion2.load( szFileName );

		if (g_bDispTwin) {
			g_clPMDModel2.setMotion( &g_clVMDMotion2, false );
			g_clPMDModel2.updateMotion( 0.0f );
			g_clPMDModel2.resetRigidBodyPos();
		}
	}
#endif
}

//--------------------------
// ŏ烂[V蒼
//--------------------------
void restartMotion( void )
{
#ifdef	_WIN32
	// ̖ڂ̏
	if (g_bDispModel) {
		g_clPMDModel.setMotion( &g_clVMDMotion, false );
		g_clPMDModel.updateMotion( 0.0f );
		g_clPMDModel.resetRigidBodyPos();
	}

	// ̖ڂ̏
	if (g_bDispTwin) {
		g_clPMDModel2.setMotion( &g_clVMDMotion2, false );
		g_clPMDModel2.updateMotion( 0.0f );
		g_clPMDModel2.resetRigidBodyPos();
	}
#endif
}

//-------------------
// FPŠv(x)
//-------------------
void calcFps( void )
{
	static int		iT1 = 0,
					iT2 = 0,
					iCnt = 0;
	static float	fFps = 0.0f;

	iT1 = glutGet( GLUT_ELAPSED_TIME );
	if( iT1 - iT2 >= 1000 )
	{
		fFps = (1000.0f * iCnt) / (float)(iT1 - iT2);
		iT2 = iT1;
		iCnt = 0;
	}
	iCnt++; 

	if( g_bDispFPS )
	{
		char	szStr[16];
		sprintf( szStr, "%6.2f", fFps );
		drawString( szStr, 8, 8 );
	}
}

//-------------------
// ݂̃fr[s`
//-------------------
void dispMatModelView( void )
{
	if( g_bDispMatrix )
	{
		char	szStr[256];
		sprintf( szStr, "%6.2f %6.2f %6.2f %6.2f", g_dMatModelView[ 0], g_dMatModelView[ 4], g_dMatModelView[ 8], g_dMatModelView[12] );
		drawString( szStr, 8, 20 );
		sprintf( szStr, "%6.2f %6.2f %6.2f %6.2f", g_dMatModelView[ 1], g_dMatModelView[ 5], g_dMatModelView[ 9], g_dMatModelView[13] );
		drawString( szStr, 8, 32 );
		sprintf( szStr, "%6.2f %6.2f %6.2f %6.2f", g_dMatModelView[ 2], g_dMatModelView[ 6], g_dMatModelView[10], g_dMatModelView[14] );
		drawString( szStr, 8, 44 );
		sprintf( szStr, "%6.2f %6.2f %6.2f %6.2f", g_dMatModelView[ 3], g_dMatModelView[ 7], g_dMatModelView[11], g_dMatModelView[15] );
		drawString( szStr, 8, 56 );
	}
}

//------------
// `
//------------
void drawString( const char *szStr, int iPosX, int iPosY )
{
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_TEXTURE_2D );

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
		glLoadIdentity();
		gluOrtho2D( 0.0f, (float)glutGet( GLUT_WINDOW_WIDTH ), 0.0f, (float)glutGet( GLUT_WINDOW_HEIGHT ) );

		glMatrixMode( GL_MODELVIEW );
		glPushMatrix();
			glLoadIdentity();

//			glColor4f( 1.0f, 1.0f, 1.0f, 1.0f ); // 
			glColor4f( 0.0f, 0.0f, 0.0f, 1.0f ); // 

			glRasterPos2i( iPosX, glutGet( GLUT_WINDOW_HEIGHT ) - (iPosY + 15) );

			while( *szStr )
			{
				glutBitmapCharacter( GLUT_BITMAP_9_BY_15, *szStr );
				szStr++;
 			}

			glMatrixMode( GL_PROJECTION );
		glPopMatrix();
		glMatrixMode( GL_MODELVIEW );
	glPopMatrix();

	glEnable( GL_DEPTH_TEST );
	glEnable( GL_LIGHTING );
}
